.sponsors {
    display: flex;
    justify-content: left;
    flex-wrap: wrap;
    align-items: center;
    margin: 1rem 0;
}

.sponsors > div {
    text-align: center;
    width: 25%;
    padding-bottom: 20px;
}

.sponsors span {
    display: block;
}
